@@ -83,12 +83,12 @@ double spacing = 1.;		// stepsize for equally spaced generated values
 double symbol_size = .01;	// index of symbol drawn at each point.
 double up = .1;			// the fractional margin above the plot.
 double width = .8;		// fraction width of the plot.
-double x_lower_limit = HUGE;	// HUGE means get it from the data
+double x_lower_limit = DBL_MAX;	// DBL_MAX means get it from the data
 double x_margin = 0.0;		// fractional margin between data and box
-double x_upper_limit = HUGE;	// HUGE means get it from the data
-double y_lower_limit = HUGE;	// HUGE means get it from the data
+double x_upper_limit = DBL_MAX;	// DBL_MAX means get it from the data
+double y_lower_limit = DBL_MAX;	// DBL_MAX means get it from the data
 double y_margin = 0.05;		// fractional margin between data and box
-double y_upper_limit = HUGE;	// HUGE means get it from the data
+double y_upper_limit = DBL_MAX;	// DBL_MAX means get it from the data
 int abcissa_flag = 0;		// nonzero means generate x axiz values
 int break_flag = 0;		// break the line whenever x decreases.
 int debug_flag = 0;		// verbose debugging output.
@@ -190,11 +190,12 @@ inline int py (double t)	// transform fractional x to plot x
 
 inline double clip (double t)
 {
-  return (t >? 0.) <? 1.;
+  double tmp = ( t > 0.0) ? t : 0.0;
+  return (tmp < 1.0) ? tmp : 1.0;
 }
 
 // uppper and lower bounds on the data
-double xmin = HUGE, ymin = HUGE, xmax = -HUGE, ymax = -HUGE;
+double xmin = DBL_MAX, ymin = DBL_MAX, xmax = -DBL_MAX, ymax = -DBL_MAX;
 double log_xmin, log_ymin, log_xmax, log_ymax;
 
 inline double fx (double t)	// transform data x to fractional x
@@ -238,7 +239,7 @@ main (int argc, char **argv)
       case 'L':	switch_style++; break;
       case 'M':
 	if ('x' == getopt.first_char())
-	  {getopt.optind++; getopt.next_arg (x_margin);} break;
+	  {getopt.optind++; getopt.next_arg (x_margin); break;}
 	if ('y' == getopt.first_char())
 	  {getopt.optind++; getopt.next_arg (y_margin);} break;
       case 'N':	getopt.next_arg (no_of_ticks); break;
@@ -387,10 +388,10 @@ main (int argc, char **argv)
     }
   
   // use limits specified on the command line if present.
-  if (x_lower_limit != HUGE) xmin = x_lower_limit;
-  if (y_lower_limit != HUGE) ymin = y_lower_limit;
-  if (x_upper_limit != HUGE) xmax = x_upper_limit;
-  if (y_upper_limit != HUGE) ymax = y_upper_limit;
+  if (x_lower_limit != DBL_MAX) xmin = x_lower_limit;
+  if (y_lower_limit != DBL_MAX) ymin = y_lower_limit;
+  if (x_upper_limit != DBL_MAX) xmax = x_upper_limit;
+  if (y_upper_limit != DBL_MAX) ymax = y_upper_limit;
   
   // make sure that 0 is not in range if we are using a log scale.
   if (   (x_log_scale
@@ -463,7 +464,7 @@ main (int argc, char **argv)
 				      .5 * char_width * strlen (tick_label))),
 	     py ((x_label_on_top ? 1. : (extended_plot_format ? 0 : -1.
 					 * char_height))
-		 + (x_label_on_top ? 1. : -1.) * (0. >? size_of_ticks)));
+		 + (x_label_on_top ? 1. : -1.) * ((size_of_ticks < 0.) ? 0. : size_of_ticks)));
 	  if (extended_plot_format)
 	    plot_file.alabel (CENTER_JUSTIFY, x_label_on_top ? BOTTOM_FLUSH
 			      : TOP_FLUSH, tick_label);
@@ -488,7 +489,7 @@ main (int argc, char **argv)
 			    .5 * char_width * x_label.length ()))),
 		 py ((x_label_on_top ? 1. : (extended_plot_format ? 0 : -1.
 					     * char_height)
-		      + (x_label_on_top ? 1. : -1.) * (0. >? size_of_ticks))));
+		      + (x_label_on_top ? 1. : -1.) * ((size_of_ticks < 0.) ? 0. : size_of_ticks))));
 	      if (extended_plot_format)
 		plot_file.alabel (CENTER_JUSTIFY, x_label_on_top ? BOTTOM_FLUSH
 				  : TOP_FLUSH, x_label);
@@ -520,7 +521,7 @@ main (int argc, char **argv)
 	  plot_file.move
 	    (px ((y_label_on_right ? 1. : (extended_plot_format ? 0 : -1.
 					   * char_width * strlen(tick_label)))
-		 + (y_label_on_right ? 1. : -1.) * (0. >? size_of_ticks)),
+		 + (y_label_on_right ? 1. : -1.) * ((size_of_ticks < 0.) ? 0. : size_of_ticks)),
 	     py (fy (y_tick_value) - .5
 		 * (extended_plot_format ? 0 : char_height)));
 	  if (extended_plot_format)
@@ -545,7 +546,7 @@ main (int argc, char **argv)
 	      plot_file.move
 		(px ((y_label_on_right ? 1. : (extended_plot_format ? 0 : -1.
 					       * char_width * strlen (y_label)))
-		     + (y_label_on_right ? 1. : -1.) * (0. >? size_of_ticks)),
+		     + (y_label_on_right ? 1. : -1.) * ((size_of_ticks < 0.) ? 0. : size_of_ticks)),
 		 py (fy (y_tick_value + y_tick / 2.)));
 	      if (extended_plot_format)
 		plot_file.alabel (y_label_on_right ? LEFT_JUSTIFY
