@@ -49,7 +49,7 @@ class indirectbuf : public streambuf {
 	{ return _get_stream ? _get_stream : lookup_stream(ios::in); }
     streambuf *put_stream()
 	{ return _put_stream ? _put_stream : lookup_stream(ios::out); }
-    virtual streambuf *lookup_stream(int mode) { return NULL; } // ERROR!
+    virtual streambuf *lookup_stream(int/*mode*/) { return NULL; } // ERROR!
     indirectbuf(streambuf *get=NULL, streambuf *put=NULL, int delete_mode=0);
     virtual ~indirectbuf();
     virtual int xsputn(const char* s, int n);
