@@ -31,7 +31,11 @@
 #define EOF (-1)
 #endif
 #ifndef NULL
+#ifdef __GNUC__
 #define NULL ((void*)0)
+#else
+#define NULL (0)
+#endif
 #endif
 
 class ostream; class streambuf; class backupbuf;
@@ -96,6 +100,7 @@ class ios : public _ios_fields {
     typedef __fmtflags fmtflags;
     typedef int iostate;
     typedef int openmode;
+    typedef int streamsize;
     enum io_state {
 	goodbit = _IOS_GOOD,
 	eofbit = _IOS_EOF,
@@ -116,7 +121,7 @@ class ios : public _ios_fields {
 	   dec=020, oct=040, hex=0100,
 	   showbase=0200, showpoint=0400, uppercase=01000, showpos=02000,
 	   scientific=04000, fixed=010000, unitbuf=020000, stdio=040000,
-	   dont_close=0x80000000 //Don't delete streambuf on stream destruction
+	   dont_close=0100000 //Don't delete streambuf on stream destruction
 	   };
     enum { // Masks.
 	basefield=dec+oct+hex,
@@ -172,6 +177,8 @@ class ios : public _ios_fields {
 	if (_state & _exceptions) _throw_failure(); }
     void set(iostate flag) { _state |= flag;
 	if (_state & _exceptions) _throw_failure(); }
+    void setstate(iostate flag) { _state |= flag; // ANSI
+	if (_state & _exceptions) _throw_failure(); }
     int good() const { return _state == 0; }
     int eof() const { return _state & ios::eofbit; }
     int fail() const { return _state & (ios::badbit|ios::failbit); }
@@ -179,10 +186,10 @@ class ios : public _ios_fields {
     iostate rdstate() const { return _state; }
     operator void*() const { return fail() ? (void*)0 : (void*)(-1); }
     int operator!() const { return fail(); }
-    iostate exception(iostate enable) {
-	iostate old = _exceptions; _exceptions = enable;
-	if (_state & _exceptions) _throw_failure();
-	return old; }
+    iostate exceptions() const { return _exceptions; }
+    void exceptions(iostate enable) {
+	_exceptions = enable;
+	if (_state & _exceptions) _throw_failure(); }
 
     static int sync_with_stdio(int on);
     static void sync_with_stdio() { sync_with_stdio(1); }
@@ -195,6 +202,12 @@ class ios : public _ios_fields {
     int writable();
 #endif
 
+    // Used to initialize standard streams. Not needed in this implementation.
+    class Init {
+    public:
+      Init () { }
+    };
+
   protected:
     ios(streambuf* sb = 0, ostream* tie = 0);
     virtual ~ios();
@@ -286,7 +299,7 @@ struct __streambuf {
 
 extern unsigned __adjust_column(unsigned start, const char *line, int count);
 
-struct streambuf : private __streambuf {
+struct streambuf : public __streambuf {
     friend class ios;
     friend class istream;
     friend class ostream;
@@ -391,7 +404,7 @@ struct streambuf : private __streambuf {
 	else return *(unsigned char*)_gptr; }
     int snextc() {
 	if (_gptr >= _egptr && __underflow(this) == EOF) return EOF;
-	return _gptr++, sgetc(); }
+	else return _gptr++, sgetc(); }
     int sputc(int c) {
 	if (_pptr >= _epptr) return __overflow(this, (unsigned char)c);
 	else return *_pptr++ = c, (unsigned char)c; }
@@ -427,7 +440,7 @@ class backupbuf : public streambuf {
 struct __file_fields {
     short _fileno;
     int _blksize;
-    _G_fpos_t _offset;
+    _G_off_t _offset;
 //    char* _save_gptr;  char* _save_egptr;
 };
 
@@ -471,9 +484,9 @@ class filebuf : public backupbuf {
     virtual int sys_close();
 };
 
-inline ios::ios(streambuf* sb /* = 0 */, ostream* tie /* = 0 */) {
+inline ios::ios(streambuf* sb /* = 0 */, ostream* tie_to /* = 0 */) {
 		_state = sb ? ios::goodbit : ios::badbit; _exceptions=0;
-		_strbuf=sb; _tie = tie; _width=0; _fill=' ';
+		_strbuf=sb; _tie = tie_to; _width=0; _fill=' ';
 		_flags=ios::skipws|ios::dec; _precision=6; }
 inline ios::~ios() {
     if (!(_flags & (unsigned int)ios::dont_close)) delete _strbuf; }
