@@ -978,7 +978,7 @@ int BitString::next(int p, unsigned int b) const
   }
 }
 
-int BitString::previous(int p, unsigned int b) const
+int BitString::prev(int p, unsigned int b) const
 {
   if (--p < 0)
     return -1;
