@@ -468,7 +468,7 @@ double ratio(const Integer& num, const Integer& den)
 {
   Integer q, r;
   divide(num, den, q, r);
-  double d1 = double(q);
+  double d1 = q.as_double();
  
   if (d1 >= DBL_MAX || d1 <= -DBL_MAX || sign(r) == 0)
     return d1;
@@ -1822,13 +1822,9 @@ void (setbit)(Integer& x, long b)
   {
     int bw = (unsigned long)b / I_SHIFT;
     int sw = (unsigned long)b % I_SHIFT;
-    if (x.rep == 0)
-      x.rep = Icalloc(0, bw + 1);
-    else if (x.rep->len < bw)
-    {
-      int xl = x.rep->len;
+    int xl = x.rep ? x.rep->len : 0;
+    if (xl <= bw)
       x.rep = Iresize(x.rep, calc_len(xl, bw+1, 0));
-    }
     x.rep->s[bw] |= (1 << sw);
     Icheck(x.rep);
   }
@@ -1837,17 +1833,16 @@ void (setbit)(Integer& x, long b)
 void clearbit(Integer& x, long b)
 {
   if (b >= 0)
-  {
-    int bw = (unsigned long)b / I_SHIFT;
-    int sw = (unsigned long)b % I_SHIFT;
-    if (x.rep == 0)
-      x.rep = Icalloc(0, bw + 1);
-    else if (x.rep->len < bw)
     {
-      int xl = x.rep->len;
-      x.rep = Iresize(x.rep, calc_len(xl, bw+1, 0));
-    }
-    x.rep->s[bw] &= ~(1 << sw);
+      if (x.rep == 0)
+	x.rep = &_ZeroRep;
+      else
+	{
+	  int bw = (unsigned long)b / I_SHIFT;
+	  int sw = (unsigned long)b % I_SHIFT;
+	  if (x.rep->len > bw)
+	    x.rep->s[bw] &= ~(1 << sw);
+	}
     Icheck(x.rep);
   }
 }
@@ -2308,14 +2303,14 @@ istream& operator >> (istream& s, Integer& y)
 #else
   if (!s.ipfx(0))
   {
-    s.set(ios::failbit);
+    s.clear(ios::failbit|s.rdstate());
     return s;
   }
 #endif
   s >> ws;
   if (!s.good())
   {
-    s.set(ios::failbit);
+    s.clear(ios::failbit|s.rdstate());
     return s;
   }
   
@@ -2395,7 +2390,7 @@ istream& operator >> (istream& s, Integer& y)
   if (s.good())
     s.putback(ch);
   if (!got_one)
-    s.set(ios::failbit);
+    s.clear(ios::failbit|s.rdstate());
 
   if (sgn == '-')
     y.negate();
