@@ -28,9 +28,9 @@ Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <new.h>
 #include <builtin.h>
 
-// extern "C" {
+extern "C" {
 #include <regex.h>
-// }
+}
 
 #include <Regex.h>
 
