@@ -19,8 +19,6 @@ License along with libiberty; see the file COPYING.LIB.  If
 not, write to the Free Software Foundation, Inc., 675 Mass Ave,
 Cambridge, MA 02139, USA.  */
 
-/* This is for g++ 1.95.03 (November 13 version).  */
-
 /* This file exports two functions; cplus_mangle_opname and cplus_demangle.
 
    This file imports xmalloc and xrealloc, which are like malloc and
@@ -34,8 +32,10 @@ Cambridge, MA 02139, USA.  */
 #include <string.h>
 #include <stdio.h>
 
-extern char *xmalloc ();
-extern char *xrealloc ();
+extern char *xmalloc PARAMS((long));
+extern char *xrealloc PARAMS((PTR, long));
+extern char *strstr PARAMS ((const char *, const char *));
+extern void free PARAMS((PTR));
 
 /* In order to allow a single demangler executable to demangle strings
    using various common values of CPLUS_MARKER, as well as any specific
@@ -92,82 +92,82 @@ static CONST struct optable
   CONST char *out;
   int flags;
 } optable[] = {
-  "nw",		  " new",	DMGL_ANSI,	/* new (1.92,	 ansi) */
-  "dl",		  " delete",	DMGL_ANSI,	/* new (1.92,	 ansi) */
-  "new",	  " new",	0,		/* old (1.91,	 and 1.x) */
-  "delete",	  " delete",	0,		/* old (1.91,	 and 1.x) */
-  "as",		  "=",		DMGL_ANSI,	/* ansi */
-  "ne",		  "!=",		DMGL_ANSI,	/* old, ansi */
-  "eq",		  "==",		DMGL_ANSI,	/* old,	ansi */
-  "ge",		  ">=",		DMGL_ANSI,	/* old,	ansi */
-  "gt",		  ">",		DMGL_ANSI,	/* old,	ansi */
-  "le",		  "<=",		DMGL_ANSI,	/* old,	ansi */
-  "lt",		  "<",		DMGL_ANSI,	/* old,	ansi */
-  "plus",	  "+",		0,		/* old */
-  "pl",		  "+",		DMGL_ANSI,	/* ansi */
-  "apl",	  "+=",		DMGL_ANSI,	/* ansi */
-  "minus",	  "-",		0,		/* old */
-  "mi",		  "-",		DMGL_ANSI,	/* ansi */
-  "ami",	  "-=",		DMGL_ANSI,	/* ansi */
-  "mult",	  "*",		0,		/* old */
-  "ml",		  "*",		DMGL_ANSI,	/* ansi */
-  "amu",	  "*=",		DMGL_ANSI,	/* ansi (ARM/Lucid) */
-  "aml",	  "*=",		DMGL_ANSI,	/* ansi (GNU/g++) */
-  "convert",	  "+",		0,		/* old (unary +) */
-  "negate",	  "-",		0,		/* old (unary -) */
-  "trunc_mod",	  "%",		0,		/* old */
-  "md",		  "%",		DMGL_ANSI,	/* ansi */
-  "amd",	  "%=",		DMGL_ANSI,	/* ansi */
-  "trunc_div",	  "/",		0,		/* old */
-  "dv",		  "/",		DMGL_ANSI,	/* ansi */
-  "adv",	  "/=",		DMGL_ANSI,	/* ansi */
-  "truth_andif",  "&&",		0,		/* old */
-  "aa",		  "&&",		DMGL_ANSI,	/* ansi */
-  "truth_orif",	  "||",		0,		/* old */
-  "oo",		  "||",		DMGL_ANSI,	/* ansi */
-  "truth_not",	  "!",		0,		/* old */
-  "nt",		  "!",		DMGL_ANSI,	/* ansi */
-  "postincrement","++",		0,		/* old */
-  "pp",		  "++",		DMGL_ANSI,	/* ansi */
-  "postdecrement","--",		0,		/* old */
-  "mm",		  "--",		DMGL_ANSI,	/* ansi */
-  "bit_ior",	  "|",		0,		/* old */
-  "or",		  "|",		DMGL_ANSI,	/* ansi */
-  "aor",	  "|=",		DMGL_ANSI,	/* ansi */
-  "bit_xor",	  "^",		0,		/* old */
-  "er",		  "^",		DMGL_ANSI,	/* ansi */
-  "aer",	  "^=",		DMGL_ANSI,	/* ansi */
-  "bit_and",	  "&",		0,		/* old */
-  "ad",		  "&",		DMGL_ANSI,	/* ansi */
-  "aad",	  "&=",		DMGL_ANSI,	/* ansi */
-  "bit_not",	  "~",		0,		/* old */
-  "co",		  "~",		DMGL_ANSI,	/* ansi */
-  "call",	  "()",		0,		/* old */
-  "cl",		  "()",		DMGL_ANSI,	/* ansi */
-  "alshift",	  "<<",		0,		/* old */
-  "ls",		  "<<",		DMGL_ANSI,	/* ansi */
-  "als",	  "<<=",	DMGL_ANSI,	/* ansi */
-  "arshift",	  ">>",		0,		/* old */
-  "rs",		  ">>",		DMGL_ANSI,	/* ansi */
-  "ars",	  ">>=",	DMGL_ANSI,	/* ansi */
-  "component",	  "->",		0,		/* old */
-  "pt",		  "->",		DMGL_ANSI,	/* ansi; Lucid C++ form */
-  "rf",		  "->",		DMGL_ANSI,	/* ansi; ARM/GNU form */
-  "indirect",	  "*",		0,		/* old */
-  "method_call",  "->()",	0,		/* old */
-  "addr",	  "&",		0,		/* old (unary &) */
-  "array",	  "[]",		0,		/* old */
-  "vc",		  "[]",		DMGL_ANSI,	/* ansi */
-  "compound",	  ", ",		0,		/* old */
-  "cm",		  ", ",		DMGL_ANSI,	/* ansi */
-  "cond",	  "?:",		0,		/* old */
-  "cn",		  "?:",		DMGL_ANSI,	/* psuedo-ansi */
-  "max",	  ">?",		0,		/* old */
-  "mx",		  ">?",		DMGL_ANSI,	/* psuedo-ansi */
-  "min",	  "<?",		0,		/* old */
-  "mn",		  "<?",		DMGL_ANSI,	/* psuedo-ansi */
-  "nop",	  "",		0,		/* old (for operator=) */
-  "rm",		  "->*",	DMGL_ANSI,	/* ansi */
+  {"nw",	  " new",	DMGL_ANSI},	/* new (1.92,	 ansi) */
+  {"dl",	  " delete",	DMGL_ANSI},	/* new (1.92,	 ansi) */
+  {"new",	  " new",	0},		/* old (1.91,	 and 1.x) */
+  {"delete",	  " delete",	0},		/* old (1.91,	 and 1.x) */
+  {"as",	  "=",		DMGL_ANSI},	/* ansi */
+  {"ne",	  "!=",		DMGL_ANSI},	/* old, ansi */
+  {"eq",	  "==",		DMGL_ANSI},	/* old,	ansi */
+  {"ge",	  ">=",		DMGL_ANSI},	/* old,	ansi */
+  {"gt",	  ">",		DMGL_ANSI},	/* old,	ansi */
+  {"le",	  "<=",		DMGL_ANSI},	/* old,	ansi */
+  {"lt",	  "<",		DMGL_ANSI},	/* old,	ansi */
+  {"plus",	  "+",		0},		/* old */
+  {"pl",	  "+",		DMGL_ANSI},	/* ansi */
+  {"apl",	  "+=",		DMGL_ANSI},	/* ansi */
+  {"minus",	  "-",		0},		/* old */
+  {"mi",	  "-",		DMGL_ANSI},	/* ansi */
+  {"ami",	  "-=",		DMGL_ANSI},	/* ansi */
+  {"mult",	  "*",		0},		/* old */
+  {"ml",	  "*",		DMGL_ANSI},	/* ansi */
+  {"amu",	  "*=",		DMGL_ANSI},	/* ansi (ARM/Lucid) */
+  {"aml",	  "*=",		DMGL_ANSI},	/* ansi (GNU/g++) */
+  {"convert",	  "+",		0},		/* old (unary +) */
+  {"negate",	  "-",		0},		/* old (unary -) */
+  {"trunc_mod",	  "%",		0},		/* old */
+  {"md",	  "%",		DMGL_ANSI},	/* ansi */
+  {"amd",	  "%=",		DMGL_ANSI},	/* ansi */
+  {"trunc_div",	  "/",		0},		/* old */
+  {"dv",	  "/",		DMGL_ANSI},	/* ansi */
+  {"adv",	  "/=",		DMGL_ANSI},	/* ansi */
+  {"truth_andif", "&&",		0},		/* old */
+  {"aa",	  "&&",		DMGL_ANSI},	/* ansi */
+  {"truth_orif",  "||",		0},		/* old */
+  {"oo",	  "||",		DMGL_ANSI},	/* ansi */
+  {"truth_not",	  "!",		0},		/* old */
+  {"nt",	  "!",		DMGL_ANSI},	/* ansi */
+  {"postincrement","++",	0},		/* old */
+  {"pp",	  "++",		DMGL_ANSI},	/* ansi */
+  {"postdecrement","--",	0},		/* old */
+  {"mm",	  "--",		DMGL_ANSI},	/* ansi */
+  {"bit_ior",	  "|",		0},		/* old */
+  {"or",	  "|",		DMGL_ANSI},	/* ansi */
+  {"aor",	  "|=",		DMGL_ANSI},	/* ansi */
+  {"bit_xor",	  "^",		0},		/* old */
+  {"er",	  "^",		DMGL_ANSI},	/* ansi */
+  {"aer",	  "^=",		DMGL_ANSI},	/* ansi */
+  {"bit_and",	  "&",		0},		/* old */
+  {"ad",	  "&",		DMGL_ANSI},	/* ansi */
+  {"aad",	  "&=",		DMGL_ANSI},	/* ansi */
+  {"bit_not",	  "~",		0},		/* old */
+  {"co",	  "~",		DMGL_ANSI},	/* ansi */
+  {"call",	  "()",		0},		/* old */
+  {"cl",	  "()",		DMGL_ANSI},	/* ansi */
+  {"alshift",	  "<<",		0},		/* old */
+  {"ls",	  "<<",		DMGL_ANSI},	/* ansi */
+  {"als",	  "<<=",	DMGL_ANSI},	/* ansi */
+  {"arshift",	  ">>",		0},		/* old */
+  {"rs",	  ">>",		DMGL_ANSI},	/* ansi */
+  {"ars",	  ">>=",	DMGL_ANSI},	/* ansi */
+  {"component",	  "->",		0},		/* old */
+  {"pt",	  "->",		DMGL_ANSI},	/* ansi; Lucid C++ form */
+  {"rf",	  "->",		DMGL_ANSI},	/* ansi; ARM/GNU form */
+  {"indirect",	  "*",		0},		/* old */
+  {"method_call",  "->()",	0},		/* old */
+  {"addr",	  "&",		0},		/* old (unary &) */
+  {"array",	  "[]",		0},		/* old */
+  {"vc",	  "[]",		DMGL_ANSI},	/* ansi */
+  {"compound",	  ", ",		0},		/* old */
+  {"cm",	  ", ",		DMGL_ANSI},	/* ansi */
+  {"cond",	  "?:",		0},		/* old */
+  {"cn",	  "?:",		DMGL_ANSI},	/* psuedo-ansi */
+  {"max",	  ">?",		0},		/* old */
+  {"mx",	  ">?",		DMGL_ANSI},	/* psuedo-ansi */
+  {"min",	  "<?",		0},		/* old */
+  {"mn",	  "<?",		DMGL_ANSI},	/* psuedo-ansi */
+  {"nop",	  "",		0},		/* old (for operator=) */
+  {"rm",	  "->*",	DMGL_ANSI}	/* ansi */
 };
 
 
@@ -184,7 +184,7 @@ typedef struct string		/* Beware: these aren't required to be */
 #define APPEND_BLANK(str)	{if (!STRING_EMPTY(str)) \
 				   string_append(str, " ");}
 
-#define ARM_VTABLE_STRING "__vtbl__"	/* Lucid/cfront virtual table prefix */
+#define ARM_VTABLE_STRING "__vtbl__"	/* Lucid/ARM virtual table prefix */
 #define ARM_VTABLE_STRLEN 8		/* strlen (ARM_VTABLE_STRING) */
 
 /* Prototypes for local functions */
@@ -202,7 +202,7 @@ demangle_template PARAMS ((struct work_stuff *work, CONST char **, string *));
 
 static int
 demangle_qualified PARAMS ((struct work_stuff *, CONST char **, string *,
-			    int));
+			    int, int));
 
 static int
 demangle_class PARAMS ((struct work_stuff *, CONST char **, string *));
@@ -220,7 +220,7 @@ static int
 gnu_special PARAMS ((struct work_stuff *, CONST char **, string *));
 
 static int
-cfront_special PARAMS ((struct work_stuff *, CONST char **, string *));
+arm_special PARAMS ((struct work_stuff *, CONST char **, string *));
 
 static void
 string_need PARAMS ((string *, int));
@@ -279,13 +279,13 @@ remember_type PARAMS ((struct work_stuff *, CONST char *, int));
 static void
 forget_types PARAMS ((struct work_stuff *));
 
-#if 0
 static void
 string_prepends PARAMS ((string *, string *));
-#endif
 
 /*  Translate count to integer, consuming tokens in the process.
-    Conversion terminates on the first non-digit character. */
+    Conversion terminates on the first non-digit character.
+    Trying to consume something that isn't a count results in
+    no consumption of input and a return of 0. */
 
 static int
 consume_count (type)
@@ -293,12 +293,12 @@ consume_count (type)
 {
     int count = 0;
 
-    do
+    while (isdigit (**type))
       {
 	count *= 10;
 	count += **type - '0';
 	(*type)++;
-      } while (isdigit (**type));
+      }
     return (count);
 }
 
@@ -419,7 +419,6 @@ mop_up (work, declp, success)
      string *declp;
      int success;
 {
-  int i;
   char *demangled = NULL;
 
   /* Discard the remembered types, if any. */
@@ -485,7 +484,7 @@ demangle_signature (work, mangled, declp)
   int success = 1;
   int func_done = 0;
   int expect_func = 0;
-  CONST char *oldmangled;
+  CONST char *oldmangled = NULL;
 
   while (success && (**mangled != '\0'))
     {
@@ -493,7 +492,7 @@ demangle_signature (work, mangled, declp)
 	{
 	  case 'Q':
 	    oldmangled = *mangled;
-	    success = demangle_qualified (work, mangled, declp, 1);
+	    success = demangle_qualified (work, mangled, declp, 1, 0);
 	    if (success)
 	      {
 		remember_type (work, oldmangled, *mangled - oldmangled);
@@ -502,23 +501,35 @@ demangle_signature (work, mangled, declp)
 	      {
 		expect_func = 1;
 	      }
+	    oldmangled = NULL;
 	    break;
 	  
 	  case 'S':
 	    /* Static member function */
+	    if (oldmangled == NULL)
+	      {
+		oldmangled = *mangled;
+	      }
 	    (*mangled)++;
 	    work -> static_type = 1;
 	    break;
 
 	  case 'C':
 	    /* a const member function */
+	    if (oldmangled == NULL)
+	      {
+		oldmangled = *mangled;
+	      }
 	    (*mangled)++;
 	    work -> const_type = 1;
 	    break;
 	  
 	  case '0': case '1': case '2': case '3': case '4':
 	  case '5': case '6': case '7': case '8': case '9':
-	    oldmangled = *mangled;
+	    if (oldmangled == NULL)
+	      {
+		oldmangled = *mangled;
+	      }
 	    success = demangle_class (work, mangled, declp);
 	    if (success)
 	      {
@@ -528,6 +539,7 @@ demangle_signature (work, mangled, declp)
 	      {
 		expect_func = 1;
 	      }
+	    oldmangled = NULL;
 	    break;
 	  
 	  case 'F':
@@ -537,15 +549,16 @@ demangle_signature (work, mangled, declp)
 	     safely just consume any 'F' at this point and be compatible
 	     with either style. */
 
+	    oldmangled = NULL;
 	    func_done = 1;
 	    (*mangled)++;
 
-	    /* For lucid/cfront style we have to forget any types we might
+	    /* For lucid/ARM style we have to forget any types we might
 	       have remembered up to this point, since they were not argument
 	       types.  GNU style considers all types seen as available for
 	       back references.  See comment in demangle_args() */
 
-	    if (LUCID_DEMANGLING || CFRONT_DEMANGLING)
+	    if (LUCID_DEMANGLING || ARM_DEMANGLING)
 	      {
 		forget_types (work);
 	      }
@@ -553,8 +566,9 @@ demangle_signature (work, mangled, declp)
 	    break;
 	  
 	  case 't':
-	    /* Template */
+	    /* G++ Template */
 	    success = demangle_template (work, mangled, declp);
+	    func_done = 1;
 	    break;
 
 	  case '_':
@@ -659,10 +673,12 @@ demangle_template (work, mangled, declp)
   int success = 0;
   int done;
   CONST char *old_p;
+  CONST char *start;
   int symbol_len;
   string temp;
 
   (*mangled)++;
+  start = *mangled;
   string_init (&tname);
   string_init (&trawname);
   /* get template name */
@@ -670,9 +686,8 @@ demangle_template (work, mangled, declp)
     {
       return (0);
     }
+  remember_type (work, start, *mangled - start + r);
   string_appendn (&tname, *mangled, r);
-  string_appendn (&trawname, *mangled, r);
-  string_appendn (&trawname, "", 1);
   *mangled += r;
   string_append (&tname, "<");
   /* get size of template parameter list */
@@ -690,11 +705,11 @@ demangle_template (work, mangled, declp)
       if (**mangled == 'Z')
 	{
 	  (*mangled)++;
+	  /* temp is initialized in do_type */
 	  success = do_type (work, mangled, &temp);
-	  string_appendn (&temp, "", 1);
 	  if (success)
 	    {
-	      string_append (&tname, temp.b);
+	      string_appends (&tname, &temp);
 	    }
 	  string_delete(&temp);
 	  if (!success)
@@ -710,11 +725,11 @@ demangle_template (work, mangled, declp)
 	  is_real = 0;
 	  is_integral = 0;
 	  done = 0;
+	  /* temp is initialized in do_type */
 	  success = do_type (work, mangled, &temp);
-	  string_appendn (&temp, "", 1);
 	  if (success)
 	    {
-	      string_append (&tname, temp.b);
+	      string_appends (&tname, &temp);
 	    }
 	  string_delete(&temp);
 	  if (!success)
@@ -822,14 +837,16 @@ demangle_template (work, mangled, declp)
 	}
       need_comma = 1;
     }
-  string_append (&tname, ">::");
+  string_append (&tname, ">");
+  string_appends (&trawname, &tname);
+  string_append (&tname, "::");
   if (work -> destructor)
     {
       string_append (&tname, "~");
     }
   if (work -> constructor || work -> destructor)
     {
-      string_append (&tname, trawname.b);
+      string_appends (&tname, &trawname);
     }
   string_delete(&trawname);
   
@@ -839,7 +856,7 @@ demangle_template (work, mangled, declp)
     }
   else
     {
-      string_prepend (declp, tname.b);
+      string_prepends (declp, &tname);
       string_delete (&tname);
       
       if (work -> static_type)
@@ -856,6 +873,71 @@ demangle_template (work, mangled, declp)
   return (success);
 }
 
+static int
+arm_pt (work, mangled, n, anchor, args)
+     struct work_stuff *work;
+     CONST char *mangled;
+     int n;
+     CONST char **anchor, **args;
+{
+  /* ARM template? */
+  if (ARM_DEMANGLING && (*anchor = strstr(mangled, "__pt__")))
+    {
+	int len;
+        *args = *anchor + 6;
+	len = consume_count (args);
+        if (*args + len == mangled + n && **args == '_')
+	  {
+	    ++*args;
+	    return 1;
+	  }
+    }
+  return 0;
+}
+
+static int
+demangle_class_name (work, mangled, declp)
+     struct work_stuff *work;
+     CONST char **mangled;
+     string *declp;
+{
+  int n;
+  int success = 0;
+
+  n = consume_count (mangled);
+  if (strlen (*mangled) >= n)
+    {
+      CONST char *p;
+      CONST char *args;
+      CONST char *e = *mangled + n;
+      /* ARM template? */
+      if (arm_pt (work, *mangled, n, &p, &args))
+	{
+	  string arg;
+	  string_init (&arg);
+	  string_appendn (declp, *mangled, p - *mangled);
+	  string_append (declp, "<");
+	  /* should do error checking here */
+	  while (args < e) {
+	    string_clear (&arg);
+	    do_type (work, &args, &arg);
+	    string_appends (declp, &arg);
+	    string_append (declp, ",");
+	  }
+	  string_delete (&arg);
+	  --declp->p;
+	  string_append (declp, ">");
+        }
+      else
+	{
+	  string_appendn (declp, *mangled, n);
+        }
+      *mangled += n;
+      success = 1;
+    }
+  return (success);
+}
+
 /*
 
 LOCAL FUNCTION
@@ -897,15 +979,15 @@ demangle_class (work, mangled, declp)
      CONST char **mangled;
      string *declp;
 {
-  int n;
   int success = 0;
+  string class_name;
 
-  n = consume_count (mangled);
-  if (strlen (*mangled) >= n)
+  string_init (&class_name);
+  if (demangle_class_name (work, mangled, &class_name))
     {
       if (work -> constructor || work -> destructor)
 	{
-	  string_prependn (declp, *mangled, n);
+	  string_prepends (declp, &class_name);
 	  if (work -> destructor)
 	    {
 	      string_prepend (declp, "~");
@@ -913,10 +995,10 @@ demangle_class (work, mangled, declp)
 	  work -> constructor = work -> destructor = 0;
 	}
       string_prepend (declp, "::");
-      string_prependn (declp, *mangled, n);
-      *mangled += n;
+      string_prepends (declp, &class_name);
       success = 1;
     }
+  string_delete (&class_name);
   return (success);
 }
 
@@ -997,9 +1079,22 @@ demangle_prefix (work, mangled, declp)
 	  success = 0;
 	}
     }
-  else if ((scan == *mangled) && (isdigit (scan[2]) || (scan[2] == 'Q')))
+  else if ((scan == *mangled) &&
+	   (isdigit (scan[2]) || (scan[2] == 'Q') || (scan[2] == 't')))
     {
-      /* A GNU style constructor starts with "__<digit>" or "__Q". */
+      /* The ARM says nothing about the mangling of local variables.
+	 But cfront mangles local variables by prepending __<nesting_level>
+	 to them. As an extension to ARM demangling we handle this case.  */
+      if ((LUCID_DEMANGLING || ARM_DEMANGLING) && isdigit (scan[2]))
+	{
+	  *mangled = scan + 2;
+	  consume_count (mangled);
+	  string_append (declp, *mangled);
+	  *mangled += strlen (*mangled);
+	  return (1);
+	}
+
+      /* A GNU style constructor starts with "__[0-9Qt]. */
       work -> constructor = 1;
       *mangled = scan + 2;
     }
@@ -1008,8 +1103,8 @@ demangle_prefix (work, mangled, declp)
       /* Mangled name starts with "__".  Skip over any leading '_' characters,
 	 then find the next "__" that separates the prefix from the signature.
 	 */
-      if (!(CFRONT_DEMANGLING || LUCID_DEMANGLING)
-	  || (cfront_special (work, mangled, declp) == 0))
+      if (!(ARM_DEMANGLING || LUCID_DEMANGLING)
+	  || (arm_special (work, mangled, declp) == 0))
 	{
 	  while (*scan == '_')
 	    {
@@ -1064,6 +1159,8 @@ DESCRIPTION
 		_vt$foo		(foo virtual table)
 		_vt$foo$bar	(foo::bar virtual table)
 		_3foo$varname	(static data member)
+		_Q22rs2tu$vw	(static data member)
+		__t6vector1Zii	(constructor with template)
  */
 
 static int
@@ -1108,18 +1205,38 @@ gnu_special (work, mangled, declp)
       string_append (declp, " virtual table");
     }
   else if ((*mangled)[0] == '_'
-	   && isdigit ((*mangled)[1])
+	   && (strchr("0123456789Qt", (*mangled)[1]) != NULL)
 	   && (p = strpbrk (*mangled, cplus_markers)) != NULL)
     {
       /* static data member, "_3foo$varname" for example */
       (*mangled)++;
-      p++;
-      n = consume_count (mangled);
-      string_appendn (declp, *mangled, n);
-      string_append (declp, "::");
-      n = strlen (p);
-      string_appendn (declp, p, n);
-      (*mangled) = p + n;
+      switch (**mangled)
+	{
+	  case 'Q':
+	    success = demangle_qualified (work, mangled, declp, 0, 1);
+	    break;
+	  case 't':
+	    success = demangle_template (work, mangled, declp);
+	    break;
+	  default:
+	    n = consume_count (mangled);
+	    string_appendn (declp, *mangled, n);
+	    (*mangled) += n;
+	}
+      if (success && (p == *mangled))
+	{
+	  /* Consumed everything up to the cplus_marker, append the
+	     variable name. */
+	  (*mangled)++;
+	  string_append (declp, "::");
+	  n = strlen (*mangled);
+	  string_appendn (declp, *mangled, n);
+	  (*mangled) += n;
+	}
+      else
+	{
+	  success = 0;
+	}
     }
   else
     {
@@ -1132,18 +1249,18 @@ gnu_special (work, mangled, declp)
 
 LOCAL FUNCTION
 
-	cfront_special -- special handling of cfront/lucid mangled strings
+	arm_special -- special handling of ARM/lucid mangled strings
 
 SYNOPSIS
 
 	static int
-	cfront_special (struct work_stuff *work, const char **mangled,
+	arm_special (struct work_stuff *work, const char **mangled,
 			string *declp);
 
 
 DESCRIPTION
 
-	Process some special cfront style mangling forms that don't fit
+	Process some special ARM style mangling forms that don't fit
 	the normal pattern.  For example:
 
 		__vtbl__3foo		(foo virtual table)
@@ -1152,19 +1269,17 @@ DESCRIPTION
  */
 
 static int
-cfront_special (work, mangled, declp)
+arm_special (work, mangled, declp)
      struct work_stuff *work;
      CONST char **mangled;
      string *declp;
 {
   int n;
-  int i;
   int success = 1;
-  CONST char *p;
 
   if (strncmp (*mangled, ARM_VTABLE_STRING, ARM_VTABLE_STRLEN) == 0)
     {
-      /* Found a cfront style virtual table, get past ARM_VTABLE_STRING
+      /* Found a ARM style virtual table, get past ARM_VTABLE_STRING
          and create the decl.  Note that we consume the entire mangled
 	 input string, which means that demangle_signature has no work
 	 to do. */
@@ -1199,13 +1314,14 @@ SYNOPSIS
 
 	static int
 	demangle_qualified (struct work_stuff *, const char *mangled,
-			    string *result, int isfuncname);
+			    string *result, int isfuncname, int append);
 
 DESCRIPTION
 
 	Demangle a qualified name, such as "Q25Outer5Inner" which is
 	the mangled form of "Outer::Inner".  The demangled output is
-	appended to the result string.
+	prepended or appended to the result string according to the
+	state of the append flag.
 
 	If isfuncname is nonzero, then the qualified name we are building
 	is going to be used as a member function name, so if it is a
@@ -1221,22 +1337,25 @@ BUGS
  */
 
 static int
-demangle_qualified (work, mangled, result, isfuncname)
+demangle_qualified (work, mangled, result, isfuncname, append)
      struct work_stuff *work;
      CONST char **mangled;
      string *result;
      int isfuncname;
+     int append;
 {
   int qualifiers;
   int namelength;
   int success = 0;
+  string temp;
 
+  string_init (&temp);
   qualifiers = (*mangled)[1] - '0';
   if (qualifiers > 0 && qualifiers < 10)
     {
       /* Assume success until we discover otherwise.  Skip over the 'Q', the
 	 qualifier count, and any '_' between the qualifier count and the
-	 first name (cfront qualified names). */
+	 first name (ARM qualified names). */
 
       success = 1;
       if ((*mangled)[2] == '_')
@@ -1246,8 +1365,8 @@ demangle_qualified (work, mangled, result, isfuncname)
       (*mangled) += 2;
 
 
-      /* Pick off the names and append them to the result string as they
-	 are found, separated by '::'. */
+      /* Pick off the names and collect them in the temp buffer in the order
+	 in which they are found, separated by '::'. */
 
       while (qualifiers-- > 0)
 	{
@@ -1258,10 +1377,10 @@ demangle_qualified (work, mangled, result, isfuncname)
 	      success = 0;
 	      break;
 	    }
-	  string_appendn (result, *mangled, namelength);
+	  string_appendn (&temp, *mangled, namelength);
 	  if (qualifiers > 0)
 	    {
-	      string_appendn (result, "::", 2);
+	      string_appendn (&temp, "::", 2);
 	    }
 	  *mangled += namelength;
 	}
@@ -1273,14 +1392,31 @@ demangle_qualified (work, mangled, result, isfuncname)
 
       if (isfuncname && (work -> constructor || work -> destructor))
 	{
-	  string_appendn (result, "::", 2);
+	  string_appendn (&temp, "::", 2);
 	  if (work -> destructor)
 	    {
-	      string_append (result, "~");
+	      string_append (&temp, "~");
 	    }
-	  string_appendn (result, (*mangled) - namelength, namelength);
+	  string_appendn (&temp, (*mangled) - namelength, namelength);
+	}
+
+      /* Now either prepend the temp buffer to the result, or append it, 
+	 depending upon the state of the append flag. */
+
+      if (append)
+	{
+	  string_appends (result, &temp);
+	}
+      else
+	{
+	  if (!STRING_EMPTY (result))
+	    {
+	      string_appendn (&temp, "::", 2);
+	    }
+	  string_prepends (result, &temp);
 	}
     }
+  string_delete (&temp);
   return (success);
 }
 
@@ -1504,7 +1640,7 @@ do_type (work, mangled, result)
     {
       /* A qualified name, such as "Outer::Inner". */
       case 'Q':
-        success = demangle_qualified (work, mangled, result, 0);
+        success = demangle_qualified (work, mangled, result, 0, 1);
 	break;
 
       default:
@@ -1549,7 +1685,6 @@ demangle_fund_type (work, mangled, result)
 {
   int done = 0;
   int success = 1;
-  int n;
 
   /* First pick off any type qualifiers.  There can be more than one. */
 
@@ -1665,15 +1800,11 @@ demangle_fund_type (work, mangled, result)
       case '7':
       case '8':
       case '9':
-	n = consume_count (mangled);
-	if (strlen (*mangled) < n)
-	  {
-	    success = 0;
-	    break;
-	  }
 	APPEND_BLANK (result);
-	string_appendn (result, *mangled, n);
-	*mangled += n;
+	if (!demangle_class_name (work, mangled, result)) {
+	  --result->p;
+	  success = 0;
+	}
 	break;
       default:
 	success = 0;
@@ -1763,7 +1894,7 @@ forget_types (work)
    DECLP must be already initialised, usually non-empty.  It won't be freed
    on failure.
 
-   Note that g++ differs significantly from cfront and lucid style mangling
+   Note that g++ differs significantly from ARM and lucid style mangling
    with regards to references to previously seen types.  For example, given
    the source fragment:
 
@@ -1780,15 +1911,15 @@ forget_types (work)
      __3fooiRT0iT2iT2
      foo__FiR3fooiT1iT1
 
-   while lcc (and presumably cfront as well) produces:
+   while lcc (and presumably other ARM style compilers as well) produces:
 
      foo__FiR3fooT1T2T1T2
      __ct__3fooFiR3fooT1T2T1T2
 
    Note that g++ bases it's type numbers starting at zero and counts all
-   previously seen types, while lucid/cfront bases it's type numbers starting
+   previously seen types, while lucid/ARM bases it's type numbers starting
    at one and only considers types after it has seen the 'F' character
-   indicating the start of the function args.  For lucid/cfront style, we
+   indicating the start of the function args.  For lucid/ARM style, we
    account for this difference by discarding any previously seen types when
    we see the 'F' character, and subtracting one from the type number
    reference.
@@ -1838,7 +1969,7 @@ demangle_args (work, mangled, declp)
 	    {
 	      return (0);
 	    }
-	  if (LUCID_DEMANGLING || CFRONT_DEMANGLING)
+	  if (LUCID_DEMANGLING || ARM_DEMANGLING)
 	    {
 	      t--;
 	    }
@@ -1928,7 +2059,7 @@ demangle_function_name (work, mangled, declp, scan)
 
   (*mangled) = scan + 2;
 
-  if (LUCID_DEMANGLING || CFRONT_DEMANGLING)
+  if (LUCID_DEMANGLING || ARM_DEMANGLING)
     {
 
       /* See if we have an ARM style constructor or destructor operator.
@@ -2001,7 +2132,8 @@ demangle_function_name (work, mangled, declp, scan)
 	  string_delete (&type);
 	}
     }
-  else if (declp->b[2] == 'o' && declp->b[3] == 'p')
+  else if (declp->b[0] == '_' && declp->b[1] == '_'
+	  && declp->b[2] == 'o' && declp->b[3] == 'p')
     {
       /* ANSI.  */
       /* type conversion operator.  */
@@ -2173,8 +2305,6 @@ string_prepend (p, s)
     }
 }
 
-#if 0
-
 static void
 string_prepends (p, s)
      string *p, *s;
@@ -2185,8 +2315,6 @@ string_prepends (p, s)
     }
 }
 
-#endif
-
 static void
 string_prependn (p, s, n)
      string *p;
@@ -2247,7 +2375,7 @@ xmalloc (size)
 
 char *
 xrealloc (oldmem, size)
-    char * oldmem;
+    PTR oldmem;
     long size;
 {
   char * newmem;
@@ -2260,25 +2388,64 @@ xrealloc (oldmem, size)
   return (newmem);
 }
 
+#include "getopt.h"
+
+static char *program_name;
+extern char *program_version;
+
+static void
+usage (stream, status)
+     FILE *stream;
+     int status;
+{    
+  fprintf (stream, "\
+Usage: %s [-_] [-s {gnu,lucid,arm}] [--strip-underscores]\n\
+       [--format={gnu,lucid,arm}] [--help] [--version] [arg...]\n",
+	   program_name);
+  exit (status);
+}
+
+#define MBUF_SIZE 512
+char mbuffer[MBUF_SIZE];
+
+/* Defined in the automatically-generated ../binutils/underscore.c. */
+extern int prepends_underscore;
+
+int strip_underscore = 0;
+
+static struct option long_options[] = {
+  {"strip-underscores", no_argument, 0, '_'},
+  {"format", required_argument, 0, 's'},
+  {"help", no_argument, 0, 'h'},
+  {"version", no_argument, 0, 'v'},
+  {0, no_argument, 0, 0}
+};
+
+int
 main (argc, argv)
-  int argc;
-  char **argv;
+     int argc;
+     char **argv;
 {
-  char mangled_name[128];
   char *result;
   int c;
-  extern char *optarg;
-  extern int optind;
 
-  while ((c = getopt (argc, argv, "s:?")) != EOF)
+  program_name = argv[0];
+  strip_underscore = prepends_underscore;
+
+  while ((c = getopt_long (argc, argv, "_s:", long_options, (int *) 0)) != EOF)
     {
       switch (c)
 	{
 	  case '?':
-	    fprintf (stderr, "usage: demangle [-s style] [arg1 [arg2]] ...\n");
-	    fprintf (stderr, "style = { gnu, lucid, cfront }\n");
-	    fprintf (stderr, "reads args from stdin if none supplied\n");
+	    usage (stderr, 1);
+	    break;
+	  case 'h':
+	    usage (stdout, 0);
+	  case 'v':
+	    printf ("GNU %s version %s\n", program_name, program_version);
 	    exit (0);
+	  case '_':
+	    strip_underscore = 1;
 	    break;
 	  case 's':
 	    if (strcmp (optarg, "gnu") == 0)
@@ -2289,18 +2456,20 @@ main (argc, argv)
 	      {
 		current_demangling_style = lucid_demangling;
 	      }
-	    else if (strcmp (optarg, "cfront") == 0)
+	    else if (strcmp (optarg, "arm") == 0)
 	      {
-		current_demangling_style = cfront_demangling;
+		current_demangling_style = arm_demangling;
 	      }
 	    else
 	      {
-		fprintf (stderr, "unknown demangling style `%s'\n", optarg);
+		fprintf (stderr, "%s: unknown demangling style `%s'\n",
+			 program_name, optarg);
 		exit (1);
 	      }
 	    break;
 	}
     }
+
   if (optind < argc)
     {
       for ( ; optind < argc; optind++)
@@ -2310,9 +2479,36 @@ main (argc, argv)
     }
   else
     {
-      while (gets (mangled_name))
+      for (;;)
 	{
-	  demangle_it (mangled_name);
+	  int i = 0;
+	  c = getchar ();
+	  /* Try to read a label. */
+	  while (c != EOF && (isalnum(c) || c == '_' || c == '$' || c == '.'))
+	    {
+	      if (i >= MBUF_SIZE-1)
+		break;
+	      mbuffer[i++] = c;
+	      c = getchar ();
+	    }
+	  if (i > 0)
+	    {
+	      int skip_first = strip_underscore && i > 1 && mbuffer[0] == '_';
+	      mbuffer[i] = 0;
+	      
+	      result = cplus_demangle (mbuffer+skip_first,
+				       DMGL_PARAMS | DMGL_ANSI);
+	      if (result)
+		{
+		  fputs (result, stdout);
+		  free (result);
+		}
+	      else
+		fputs (mbuffer + skip_first, stdout);
+	    }
+	  if (c == EOF)
+	    break;
+	  putchar (c);
 	}
     }
 
